/*
 * Copyright (C) 2011-2021 Intel Corporation. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of Intel Corporation nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */



#include <stdlib.h>
#include <stdio.h>      /* vsnprintf */
#include <stdarg.h>
#include <string.h>


#include "TestEnclave.h"
#include "TestEnclave_t.h"  /* print_string */
#include "tSgxSSL_api.h"

#include <openssl/ec.h>
#include <openssl/bn.h>
#include <openssl/rsa.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/rand.h>


#define ADD_ENTROPY_SIZE	16
#define hmacKeyLen          20

/* HMAC SHA1 Begins*/
#define SHA1HANDSOFF (1)

#include <stdint.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include "sha.h"

void SHA1_Transform(uint32_t state[5], const uint8_t buffer[64]);

#define rol(value, bits) (((value) << (bits)) | ((value) >> (32 - (bits))))

#if defined (BYTE_ORDER) && defined(BIG_ENDIAN) && (BYTE_ORDER == BIG_ENDIAN)
#define WORDS_BIGENDIAN 1
#endif
#ifdef _BIG_ENDIAN
#define WORDS_BIGENDIAN 1
#endif


/* blk0() and blk() perform the initial expand. */
/* I got the idea of expanding during the round function from SSLeay */
/* FIXME: can we do this in an endian-proof way? */
#ifdef WORDS_BIGENDIAN
#define blk0(i) block->l[i]
#else
#define blk0(i) (block->l[i] = (rol(block->l[i],24)&0xff00ff00) \
         |(rol(block->l[i],8)&0x00ff00ff))
#endif
#define blk(i) (block->l[i&15] = rol(block->l[(i+13)&15]^block->l[(i+8)&15] \
                     ^block->l[(i+2)&15]^block->l[i&15],1))

/* (R0+R1), R2, R3, R4 are the different operations used in SHA1 */
#define R0(v, w, x, y, z, i) \
    z+=((w&(x^y))^y)+blk0(i)+0x5a827999+rol(v,5);w=rol(w,30);
#define R1(v, w, x, y, z, i) \
    z+=((w&(x^y))^y)+blk(i)+0x5a827999+rol(v,5);w=rol(w,30);
#define R2(v, w, x, y, z, i) \
    z+=(w^x^y)+blk(i)+0x6ed9eba1+rol(v,5);w=rol(w,30);
#define R3(v, w, x, y, z, i) \
    z+=(((w|x)&y)|(w&x))+blk(i)+0x8f1bbcdc+rol(v,5);w=rol(w,30);
#define R4(v, w, x, y, z, i) \
    z+=(w^x^y)+blk(i)+0xca62c1d6+rol(v,5);w=rol(w,30);


/* Hash a single 512-bit block. This is the core of the algorithm. */
void SHA1_Transform(uint32_t state[5], const uint8_t buffer[64]) {
    uint32_t a, b, c, d, e;
    typedef union {
        uint8_t c[64];
        uint32_t l[16];
    } CHAR64LONG16;
    CHAR64LONG16 *block;

#ifdef SHA1HANDSOFF
    CHAR64LONG16 workspace;
    block = &workspace;
    memcpy(block, buffer, 64);
#else
	block = (CHAR64LONG16*)buffer;
#endif

    /* Copy context->state[] to working vars */
    a = state[0];
    b = state[1];
    c = state[2];
    d = state[3];
    e = state[4];

    /* 4 rounds of 20 operations each. Loop unrolled. */
    R0(a, b, c, d, e, 0);
    R0(e, a, b, c, d, 1);
    R0(d, e, a, b, c, 2);
    R0(c, d, e, a, b, 3);
    R0(b, c, d, e, a, 4);
    R0(a, b, c, d, e, 5);
    R0(e, a, b, c, d, 6);
    R0(d, e, a, b, c, 7);
    R0(c, d, e, a, b, 8);
    R0(b, c, d, e, a, 9);
    R0(a, b, c, d, e, 10);
    R0(e, a, b, c, d, 11);
    R0(d, e, a, b, c, 12);
    R0(c, d, e, a, b, 13);
    R0(b, c, d, e, a, 14);
    R0(a, b, c, d, e, 15);
    R1(e, a, b, c, d, 16);
    R1(d, e, a, b, c, 17);
    R1(c, d, e, a, b, 18);
    R1(b, c, d, e, a, 19);
    R2(a, b, c, d, e, 20);
    R2(e, a, b, c, d, 21);
    R2(d, e, a, b, c, 22);
    R2(c, d, e, a, b, 23);
    R2(b, c, d, e, a, 24);
    R2(a, b, c, d, e, 25);
    R2(e, a, b, c, d, 26);
    R2(d, e, a, b, c, 27);
    R2(c, d, e, a, b, 28);
    R2(b, c, d, e, a, 29);
    R2(a, b, c, d, e, 30);
    R2(e, a, b, c, d, 31);
    R2(d, e, a, b, c, 32);
    R2(c, d, e, a, b, 33);
    R2(b, c, d, e, a, 34);
    R2(a, b, c, d, e, 35);
    R2(e, a, b, c, d, 36);
    R2(d, e, a, b, c, 37);
    R2(c, d, e, a, b, 38);
    R2(b, c, d, e, a, 39);
    R3(a, b, c, d, e, 40);
    R3(e, a, b, c, d, 41);
    R3(d, e, a, b, c, 42);
    R3(c, d, e, a, b, 43);
    R3(b, c, d, e, a, 44);
    R3(a, b, c, d, e, 45);
    R3(e, a, b, c, d, 46);
    R3(d, e, a, b, c, 47);
    R3(c, d, e, a, b, 48);
    R3(b, c, d, e, a, 49);
    R3(a, b, c, d, e, 50);
    R3(e, a, b, c, d, 51);
    R3(d, e, a, b, c, 52);
    R3(c, d, e, a, b, 53);
    R3(b, c, d, e, a, 54);
    R3(a, b, c, d, e, 55);
    R3(e, a, b, c, d, 56);
    R3(d, e, a, b, c, 57);
    R3(c, d, e, a, b, 58);
    R3(b, c, d, e, a, 59);
    R4(a, b, c, d, e, 60);
    R4(e, a, b, c, d, 61);
    R4(d, e, a, b, c, 62);
    R4(c, d, e, a, b, 63);
    R4(b, c, d, e, a, 64);
    R4(a, b, c, d, e, 65);
    R4(e, a, b, c, d, 66);
    R4(d, e, a, b, c, 67);
    R4(c, d, e, a, b, 68);
    R4(b, c, d, e, a, 69);
    R4(a, b, c, d, e, 70);
    R4(e, a, b, c, d, 71);
    R4(d, e, a, b, c, 72);
    R4(c, d, e, a, b, 73);
    R4(b, c, d, e, a, 74);
    R4(a, b, c, d, e, 75);
    R4(e, a, b, c, d, 76);
    R4(d, e, a, b, c, 77);
    R4(c, d, e, a, b, 78);
    R4(b, c, d, e, a, 79);

    /* Add the working vars back into context.state[] */
    state[0] += a;
    state[1] += b;
    state[2] += c;
    state[3] += d;
    state[4] += e;

    /* Wipe variables */
    a = b = c = d = e = 0;
}


/**
* Initialize new context
*
* @param context SHA1-Context
*/
void SHA1_Init(SHA1_CTX *context) {
    /* SHA1 initialization constants */
    context->state[0] = 0x67452301;
    context->state[1] = 0xefcdab89;
    context->state[2] = 0x98badcfe;
    context->state[3] = 0x10325476;
    context->state[4] = 0xc3d2e1f0;
    context->count[0] = context->count[1] = 0;
}


/**
* Run your data through this
*
* @param context SHA1-Context
* @param p       Buffer to run SHA1 on
* @param len     Number of bytes
*/
void SHA1_Update(SHA1_CTX *context, const void *p, size_t len) {
    const uint8_t *data = (unsigned char*)p;
    size_t i, j;

    j = (context->count[0] >> 3) & 63;
    if ((context->count[0] += (uint32_t) (len << 3)) < (len << 3)) {
        context->count[1]++;
    }
    context->count[1] += (uint32_t) (len >> 29);
    if ((j + len) > 63) {
        memcpy(&context->buffer[j], data, (i = 64 - j));
        SHA1_Transform(context->state, context->buffer);
        for (; i + 63 < len; i += 64) {
            SHA1_Transform(context->state, data + i);
        }
        j = 0;
    }
    else i = 0;
    memcpy(&context->buffer[j], &data[i], len - i);
}


/**
* Add padding and return the message digest
*
* @param digest  Generated message digest
* @param context SHA1-Context
*/
void SHA1_Final(uint8_t digest[SHA1_DIGEST_SIZE], SHA1_CTX *context) {
    uint32_t i;
    uint8_t finalcount[8];

    for (i = 0; i < 8; i++) {
        finalcount[i] = (uint8_t) ((context->count[(i >= 4 ? 0 : 1)]
                >> ((3 - (i & 3)) * 8)) & 255);
    }
    SHA1_Update(context, (uint8_t *) "\200", 1);
    while ((context->count[0] & 504) != 448) {
        SHA1_Update(context, (uint8_t *) "\0", 1);
    }
    SHA1_Update(context, finalcount, 8); /* Should cause SHA1_Transform */
    for (i = 0; i < SHA1_DIGEST_SIZE; i++) {
        digest[i] = (uint8_t)
                ((context->state[i >> 2] >> ((3 - (i & 3)) * 8)) & 255);
    }

    /* Wipe variables */
    i = 0;
    memset(context->buffer, 0, 64);
    memset(context->state, 0, 20);
    memset(context->count, 0, 8);
    memset(finalcount, 0, 8);    /* SWR */

#ifdef SHA1HANDSOFF  /* make SHA1Transform overwrite its own static vars */
    SHA1_Transform(context->state, context->buffer);
#endif
}

/** SHA-1 Block size */
#ifndef SHA_BLOCKSIZE
#define SHA_BLOCKSIZE   (64)
#endif


/**
* Function to compute the digest
*
* @param k   Secret key
* @param lk  Length of the key in bytes
* @param d   Data
* @param ld  Length of data in bytes
* @param out Digest output
* @param t   Size of digest output
*/
void hmac_sha1(const uint8_t *k,  /* secret key */
        size_t lk,       /* length of the key in bytes */
        const uint8_t *d,  /* data */
        size_t ld,       /* length of data in bytes */
        uint8_t *out,      /* output buffer, at least "t" bytes */
        size_t *t) {
#ifdef USE_OPENSSL

	if (!HMAC(EVP_sha1(), k, (int)lk, d, ld, out, t)) {
		ERR_clear_error();
	}
#else
    SHA_CTX ictx, octx;
    uint8_t isha[SHA_DIGEST_LENGTH], osha[SHA_DIGEST_LENGTH];
    uint8_t key[SHA_DIGEST_LENGTH];
    uint8_t buf[SHA_BLOCKSIZE];
    size_t i;

    if (lk > SHA_BLOCKSIZE) {
        SHA_CTX tctx;

        SHA1_Init(&tctx);
        SHA1_Update(&tctx, k, lk);
        SHA1_Final(key, &tctx);

        k = key;
        lk = SHA_DIGEST_LENGTH;
    }

    /**** Inner Digest ****/

    SHA1_Init(&ictx);

    /* Pad the key for inner digest */
    for (i = 0; i < lk; ++i) {
        buf[i] = k[i] ^ 0x36;
    }
    for (i = lk; i < SHA_BLOCKSIZE; ++i) {
        buf[i] = 0x36;
    }

    SHA1_Update(&ictx, buf, SHA_BLOCKSIZE);
    SHA1_Update(&ictx, d, ld);

    SHA1_Final(isha, &ictx);

    /**** Outer Digest ****/

    SHA1_Init(&octx);

    /* Pad the key for outter digest */

    for (i = 0; i < lk; ++i) {
        buf[i] = k[i] ^ 0x5c;
    }
    for (i = lk; i < SHA_BLOCKSIZE; ++i) {
        buf[i] = 0x5c;
    }

    SHA1_Update(&octx, buf, SHA_BLOCKSIZE);
    SHA1_Update(&octx, isha, SHA_DIGEST_LENGTH);

    SHA1_Final(osha, &octx);

    /* truncate and print the results */
    *t = *t > SHA_DIGEST_LENGTH ? SHA_DIGEST_LENGTH : *t;
    memcpy(out, osha, *t);
#endif
}


/* calculate rprime and N via extended Euclidean Algorithm */
BIGNUM* exgcd(BIGNUM *a, BIGNUM *b, BIGNUM *x, BIGNUM *y) 
{
    BN_dec2bn(&x, "1");
    BN_dec2bn(&y, "0");
    if(BN_is_zero(b)) {
        return a;
    }
    BN_CTX *ctx = BN_CTX_new();
    BIGNUM *new_x = BN_new();
    BIGNUM *new_y = BN_new();
    BIGNUM *new_r = BN_new();
    BIGNUM *r = BN_new();
    BIGNUM *quotient = BN_new();
    BIGNUM *tmp = BN_new();
    BIGNUM *rem = BN_new();
    BIGNUM *multi = BN_new();
    //*r = *a;
    BN_dec2bn(&new_x, "0");
    BN_dec2bn(&new_y, "1");
    BN_copy(new_r, b);
    BN_copy(r, a);
    while(!BN_is_zero(new_r)) {
        BN_div(quotient, rem, r, new_r, ctx);

        BN_copy(tmp, r);
        BN_copy(r, new_r);
        BN_mul(multi, quotient, new_r, ctx);
        BN_sub(new_r, tmp, multi);


        
        BN_copy(tmp, x);
        BN_copy(x, new_x);
        BN_mul(multi, quotient, new_x, ctx);
        BN_sub(new_x, tmp, multi);

        BN_copy(tmp, y);
        BN_copy(y, new_y);
        BN_mul(multi, quotient, new_y, ctx);
        BN_sub(new_y, tmp, multi);
    }
    BN_free(new_x);
    BN_free(new_y);
    BN_free(new_r);
    //BN_free(r);
    BN_free(quotient);
    BN_free(tmp);
    BN_free(rem);
    BN_free(multi);
    return r;
}

/* HMAC SHA1 Ends */
unsigned int counter = 0;

/**************A simple random funtion generator***********/
void get_rand(BIGNUM *rand)
{

    //char *key = (char*)malloc(hmacKeyLen + 1);
    //unsigned char *data = (unsigned char*)malloc(strlen(m));
    counter++;
    unsigned char str[21];
    unsigned char key[] = "fdskjfdlsajflsajflsa";
    //unsigned char data[] = "defghijklmndikenrvvv";
    unsigned char *result = (unsigned char*)malloc(20);
    //unsigned char result[256] = {0};
    size_t len = 20;
    //strlcpy(key, "abcdkhuejndkmjueknmd", 20); // key
    //strcpy(data, mStr); // data
    //HMAC(EVP_sha1(), key, hmacKeyLen, m, strlen(m), result, NULL);
    //PKCS5_PBKDF2_HMAC_SHA1(key, strlen(key), NULL, 0, 1000, 20, result);
    snprintf((char*)str, 20, "%d", counter);
    //printf(" counter is %s\n", str);
    hmac_sha1(key, 20, str, 20, result, &len);
    BN_bin2bn(result, 16, rand);
    /*for(int i = 0; i < 20; i++) {
        printf("%x\n",result[i]);
    }*/
    //printf("rand is %s\n", result);
    //BNprint(rand);
    //free(key);
    //free(mStr);
    //free(data);
    free(result);
}

/* Initialize some parameters */
void Init(BIGNUM *p, BIGNUM *q, BIGNUM *x, BN_CTX *ctx, BIGNUM *gx, BIGNUM *g)
{
	char *tempStr = "sdfdsfsdfsdfdsjflksjdflkjdsfl";

	/* initialize p */
	BN_dec2bn(&p, "369561186540684896360312819830761123257567356085640050367370498975930251847847");


	/* initialize q */
	BN_dec2bn(&q, "184780593270342448180156409915380561628783678042820025183685249487965125923923");

	/* calculate random number */
	//get_rand(rand, tempStr);

	/* initialize g */
	BN_hex2bn(&g, "2");

	/* initialize x */
	BN_dec2bn(&x, "114780593270342448180156409915380561628783678042820025183685249487965125923923");

	/* (g ^ x) mod p */ 
	BN_mod_exp_mont_consttime(gx, g, x, p, ctx, NULL);

	/* (y ^ r) mod p*/
	//BN_mod_exp(yr, gx, rand, p, ctx);

	/* x * r */
	//BN_mul(xr, x, rand, ctx);

}


/* calculate hash value */
void chamHash1(BIGNUM *m,  BIGNUM *p, BIGNUM *yr, BIGNUM *g, char** results)
{
    //char *hash;
    BIGNUM *chamhash = BN_new();
    BIGNUM *gm = BN_new();
    BN_CTX *ctx = BN_CTX_new(); // used internally by the bignum lib

    //char* results = NULL;

    // g ^ m
    BN_mod_exp(gm, g, m, p,ctx); // use mod exp here

    // (g^m * y^r) mod p

    //printf("gm is %s\n", BN_bn2hex(gm));
    //printf("yr is %s\n", BN_bn2hex(yr));
    //printf("p is %s\n", BN_bn2hex(p));

    
    BN_mod_mul(chamhash, gm, yr, p, ctx);

    //printf("cham hash is %s\n", BN_bn2hex(chamhash));

    *results = BN_bn2hex(chamhash);

    BN_free(chamhash);
    BN_CTX_free(ctx);
    BN_free(gm);
    
    //return results;
}

    static BIGNUM *p = NULL;
	static BIGNUM *q = NULL;
	static BIGNUM *x = NULL;
	//static BIGNUM *rand = NULL;
	static BN_CTX *ctx = NULL;
	static BIGNUM *gx = NULL;
	//static BIGNUM *yr = NULL;
	//static BIGNUM *xr = NULL;
	static BIGNUM *g = NULL;
    static BIGNUM *rPrime = NULL;
    static BIGNUM *N = NULL;
	static int flag = 0;
    int oneMegeByte = 1048576; //1M bytes
    int mLen = 16;
    int hexLen = mLen * 2;
    //int count = 0;


/* calculate r and hash */
void findColandComp(char *mmPrimeStr, char *hash, char *r, int hashSize, int rSize)
{
	// BN_CTX *ctxTemp = BN_CTX_new();
	// //BIGNUM *rPrime = BN_new();
    // BIGNUM *rem = BN_new();
    // BIGNUM *temp = BN_new();
    // BIGNUM *tempXr = BN_new();
	// BIGNUM *m = BN_new();
	// BIGNUM *mPrime = BN_new();
    // BIGNUM *rPrimeFinal = BN_new();
    // //BIGNUM *NFinal = BN_new();

	// char *rPrimeStr = NULL;
    // char *hashInSGX = NULL;

    // char mStr[hexLen * 2 + 1];
    // char mPrimeStr[hexLen * 2 + 1];

    // mStr[hexLen * 2] = '\0';
    // mPrimeStr[hexLen * 2] = '\0';
    //printf("count is %d", count++);
    //printf("cover is %d \n",strlen(mStr));
    /*printf("secret is %d \n",strlen(mmPrimeStr));
    for(int i = 0; i < 2097153; i++) {
        printf("%c", mPrimeStr[i]);
    }
    return;*/
    char *mStr = "6162636466656466616263646665646661626364666564666162636466656466";

	/* initialize the parameters only once */
	if(flag == 0) {

         p = BN_new();
         q = BN_new();
         x = BN_new();
         //rand = BN_new();
         ctx = BN_CTX_new();
         gx = BN_new();
         //yr = BN_new();
         //xr = BN_new();
         g = BN_new();
         rPrime = BN_new();
         N = BN_new();

		Init(p, q, x, ctx, gx, g);

         /* calculate rPrime_0 and N_0 */
        exgcd(x, q, rPrime, N);
		flag = 1;
		printf("enter init()\n");
	}
	
	//printf("%s\n", mStr);
    //printf("%s\n", mPrimeStr);

    /* test initialization function */
    //printf("gx is %s\n",BN_bn2hex(gx));
    //printf("xr is %s\n",BN_bn2hex(xr));
    //printf("yr is %s\n",BN_bn2hex(yr));
    /* test ends */


    for(int j = 0; j < 5 * oneMegeByte; j += hexLen){
        /***********************test start******************/
        BN_CTX *ctxTemp = BN_CTX_new();
	//BIGNUM *rPrime = BN_new();
    BIGNUM *rem = BN_new();
    BIGNUM *temp = BN_new();
    BIGNUM *tempXr = BN_new();
	BIGNUM *m = BN_new();
	BIGNUM *mPrime = BN_new();
    BIGNUM *rPrimeFinal = BN_new();
    BIGNUM *xr = BN_new();
    BIGNUM *yr = BN_new();
    BIGNUM *rand = BN_new();
    //BIGNUM *NFinal = BN_new();

	char *rPrimeStr = NULL;
    char *hashInSGX = NULL;

    //char mStr[hexLen * 2+ 1];
    char mPrimeStr[hexLen * 2+ 1];

    //mStr[hexLen * 2] = '\0';
    mPrimeStr[hexLen * 2] = '\0';
        /***********************test end***************************/


        //memcpy(mStr, mmStr, hexLen * 2);
        memcpy(mPrimeStr, mmPrimeStr, hexLen * 2);
        //printf("mStr is %d\n", strlen(mStr));
        //printf("mPrimeStr is %d\n", strlen(mPrimeStr));
        //mmStr += hexLen * 2;
        mmPrimeStr += hexLen * 2;

        

        BN_hex2bn(&m, mStr);
        BN_hex2bn(&mPrime, mPrimeStr);
        // m + xr - m' = xr' mod q

        /* calculate random number */
        get_rand(rand);

        /* x * r */
        BN_mul(xr, x, rand, ctxTemp);


        /* (y ^ r) mod p*/
        BN_mod_exp(yr, gx, rand, p, ctxTemp);

        //x * r + m
        BN_add(tempXr, m, xr);

        


        //x * r + m - m'
        BN_mod_sub(temp, tempXr, mPrime, q, ctxTemp);

        

        // find r' such that m + xr - m' = xr' mod q
        BN_mod_mul(rPrimeFinal, rPrime, temp, q, ctx);
        

        
        /* calculate hash value */
         chamHash1(m, p, yr, g, &hashInSGX);
           

        
        /* transfer rPrime to hex */
        rPrimeStr = BN_bn2hex(rPrimeFinal);
        

        
        /* copy hash to outside */
        memcpy(hash, hashInSGX, strlen(hashInSGX));
        

        
        /* copy r to outside */
        memcpy(r, rPrimeStr, strlen(rPrimeStr));
        

        hash += strlen(hashInSGX);
        
        r +=  strlen(rPrimeStr);
        

        /***************************test start**************/
        BN_CTX_free(ctxTemp);
    	BN_free(m);
    	BN_free(mPrime);
    	BN_free(rPrimeFinal);
    	BN_free(rem);
    	BN_free(temp);
    	BN_free(tempXr);
        BN_free(xr);
        BN_free(yr);
        BN_free(rand);
        free(rPrimeStr);
        free(hashInSGX);
        /**************************test end*****************/
    }


}

/* 
 * printf: 
 *   Invokes OCALL to display the enclave buffer to the terminal.
 */
void printf(const char *fmt, ...)
{
    char buf[BUFSIZ] = {'\0'};
    va_list ap;
    va_start(ap, fmt);
    vsnprintf(buf, BUFSIZ, fmt, ap);
    va_end(ap);
    uprint(buf);
}

typedef void CRYPTO_RWLOCK;

struct evp_pkey_st {
    int type;
    int save_type;
    int references;
    const EVP_PKEY_ASN1_METHOD *ameth;
    ENGINE *engine;
    union {
        char *ptr;
# ifndef OPENSSL_NO_RSA
        struct rsa_st *rsa;     /* RSA */
# endif
# ifndef OPENSSL_NO_DSA
        struct dsa_st *dsa;     /* DSA */
# endif
# ifndef OPENSSL_NO_DH
        struct dh_st *dh;       /* DH */
# endif
# ifndef OPENSSL_NO_EC
        struct ec_key_st *ec;   /* ECC */
# endif
    } pkey;
    int save_parameters;
    STACK_OF(X509_ATTRIBUTE) *attributes; /* [ 0 ] */
    CRYPTO_RWLOCK *lock;
} /* EVP_PKEY */ ;

void rsa_key_gen()
{
	BIGNUM *bn = BN_new();
	if (bn == NULL) {
		printf("BN_new failure: %ld\n", ERR_get_error());
	    return;
	}
	int ret = BN_set_word(bn, RSA_F4);
    if (!ret) {
       	printf("BN_set_word failure\n");
	    return;
	}
	
	RSA *keypair = RSA_new();
	if (keypair == NULL) {
		printf("RSA_new failure: %ld\n", ERR_get_error());
	    return;
	}
	ret = RSA_generate_key_ex(keypair, 4096, bn, NULL);
	if (!ret) {
        printf("RSA_generate_key_ex failure: %ld\n", ERR_get_error());
	    return;
	}

	EVP_PKEY *evp_pkey = EVP_PKEY_new();
	if (evp_pkey == NULL) {
		printf("EVP_PKEY_new failure: %ld\n", ERR_get_error());
		return;
	}
	EVP_PKEY_assign_RSA(evp_pkey, keypair);

	// public key - string
	int len = i2d_PublicKey(evp_pkey, NULL);
	unsigned char *buf = (unsigned char *) malloc (len + 1);
	unsigned char *tbuf = buf;
	i2d_PublicKey(evp_pkey, &tbuf);

	// print public key
	printf ("{\"public\":\"");
	int i;
	for (i = 0; i < len; i++) {
	    printf("%02x", (unsigned char) buf[i]);
	}
	printf("\"}\n");

	free(buf);

	// private key - string
	len = i2d_PrivateKey(evp_pkey, NULL);
	buf = (unsigned char *) malloc (len + 1);
	tbuf = buf;
	i2d_PrivateKey(evp_pkey, &tbuf);

	// print private key
	printf ("{\"private\":\"");
	for (i = 0; i < len; i++) {
	    printf("%02x", (unsigned char) buf[i]);
	}
	printf("\"}\n");

	free(buf);

	BN_free(bn);

	EVP_PKEY_free(evp_pkey);

	if (evp_pkey->pkey.ptr != NULL) {
	  RSA_free(keypair);
	}
}

void ec_key_gen()
{
	unsigned char entropy_buf[ADD_ENTROPY_SIZE] = {0};

	RAND_add(entropy_buf, sizeof(entropy_buf), ADD_ENTROPY_SIZE);
	RAND_seed(entropy_buf, sizeof(entropy_buf));

	EC_KEY * ec = NULL;
    int eccgroup;
    eccgroup = OBJ_txt2nid("secp384r1");
    ec = EC_KEY_new_by_curve_name(eccgroup);
    if (ec == NULL) {
    	printf("EC_KEY_new_by_curve_name failure: %ld\n", ERR_get_error());
	    return;
    }
    
	EC_KEY_set_asn1_flag(ec, OPENSSL_EC_NAMED_CURVE);

	int ret = EC_KEY_generate_key(ec);
	if (!ret) {
        printf("EC_KEY_generate_key failure\n");
	    return;
	}

	EVP_PKEY *ec_pkey = EVP_PKEY_new();
	if (ec_pkey == NULL) {
		printf("EVP_PKEY_new failure: %ld\n", ERR_get_error());
		return;
	}
	EVP_PKEY_assign_EC_KEY(ec_pkey, ec);

	// public key - string
	int len = i2d_PublicKey(ec_pkey, NULL);
	unsigned char *buf = (unsigned char *) malloc (len + 1);
	unsigned char *tbuf = buf;
	i2d_PublicKey(ec_pkey, &tbuf);

	// print public key
	printf ("{\"public\":\"");
	int i;
	for (i = 0; i < len; i++) {
	    printf("%02x", (unsigned char) buf[i]);
	}
	printf("\"}\n");

	free(buf);

	// private key - string
	len = i2d_PrivateKey(ec_pkey, NULL);
	buf = (unsigned char *) malloc (len + 1);
	tbuf = buf;
	i2d_PrivateKey(ec_pkey, &tbuf);

	// print private key
	printf ("{\"private\":\"");
	for (i = 0; i < len; i++) {
	    printf("%02x", (unsigned char) buf[i]);
	}
	printf("\"}\n");

	free(buf);

	EVP_PKEY_free(ec_pkey);
	if (ec_pkey->pkey.ptr != NULL) {
	  EC_KEY_free(ec);
	}
}


int vprintf_cb(Stream_t stream, const char * fmt, va_list arg)
{
	char buf[BUFSIZ] = {'\0'};

	int res = vsnprintf(buf, BUFSIZ, fmt, arg);
	if (res >=0) {
		sgx_status_t sgx_ret = uprint((const char *) buf);
		TEST_CHECK(sgx_ret);
	}
	return res;
}

/*
extern "C" int CRYPTO_set_mem_functions(
        void *(*m)(size_t, const char *, int),
        void *(*r)(void *, size_t, const char *, int),
        void (*f)(void *, const char *, int));
void* priv_malloc(size_t size, const char *file, int line)
{
	void* addr = malloc(size);
	
	printf("[malloc:%s:%d] size: %d, addr: %p\n", file, line, size, addr);
	
	return addr;
}
void* priv_realloc(void* old_addr, size_t new_size, const char *file, int line)
{
	void* new_addr = realloc(old_addr, new_size);
	
	printf("[realloc:%s:%d] old_addr: %p, new_size: %d, new_addr: %p\n", file, line, old_addr, new_size, new_addr);
	
	return new_addr;
}
void priv_free(void* addr, const char *file, int line)
{
	printf("[free:%s:%d] addr: %p\n", file, line, addr);
	
	free(addr);
}
*/


void t_sgxssl_call_apis()
{
    int ret = 0;
    
    printf("Start tests\n");
    
    SGXSSLSetPrintToStdoutStderrCB(vprintf_cb);

    //CRYPTO_set_mem_functions(priv_malloc, priv_realloc, priv_free);

    // Initialize SGXSSL crypto
    OPENSSL_init_crypto(0, NULL);
    
    rsa_key_gen();
    printf("test rsa_key_gen completed\n");
           
    ec_key_gen();
	printf("test ec_key_gen completed\n");
	
    ret = rsa_test();
    if (ret != 0)
    {
    	printf("test rsa_test returned error %d\n", ret);
    	exit(ret);
    }
	printf("test rsa_test completed\n");

	ret = ec_test();
	if (ret != 0)
    {
    	printf("test ec_test returned error %d\n", ret);
    	exit(ret);
    }
	printf("test ec_test completed\n");
	
	ret = ecdh_test();
	if (ret != 0)
    {
    	printf("test ecdh_test returned error %d\n", ret);
    	exit(ret);
    }
	printf("test ecdh_test completed\n");
	
	ret = ecdsa_test();
	if (ret != 0)
    {
    	printf("test ecdsar_test returned error %d\n", ret);
    	exit(ret);
    }
	printf("test ecdsa_test completed\n");

	ret = bn_test();
	if (ret != 0)
    {
    	printf("test bn_test returned error %d\n", ret);
    	exit(ret);
    }
	printf("test bn_test completed\n");

	ret = dh_test();
	if (ret != 0)
    {
    	printf("test dh_test returned error %d\n", ret);
    	exit(ret);
    }
	printf("test dh_test completed\n");

	ret = sha256_test();
	if (ret != 0)
    {
    	printf("test sha256_test returned error %d\n", ret);
    	exit(ret);
    }
	printf("test sha256_test completed\n");
	
	ret = sha1_test();
	if (ret != 0)
    {
    	printf("test sha1_test returned error %d\n", ret);
    	exit(ret);
    }
	printf("test sha1_test completed\n");
	
	ret = threads_test();
	if (ret != 0)
    {
    	printf("test threads_test returned error %d\n", ret);
    	exit(ret);
    }
	printf("test threads_test completed\n");
	
    //GM SM2 - sign and verify
    ret = ecall_sm2();
    if (ret != 0)
    {
        printf("test evp_sm2 returned error %d\n", ret);
        exit(ret);
    }
    printf("test evp_sm2 completed\n");

    //GM SM3 - compute digest of message
    ret = ecall_sm3();
    if (ret != 0)
    {
        printf("test evp_sm3 returned error %d\n", ret);
        exit(ret);
    }
    printf("test evp_sm3 completed\n");

    //GM SM4 - cbc encrypt and decrypt
    ret = ecall_sm4_cbc();
    if (ret != 0)
    {
        printf("test evp_sm4_cbc returned error %d\n", ret);
        exit(ret);
    }
    printf("test evp_sm4_cbc completed\n");

    //GM SM4 - ctr encrypt and decrypt
    ret = ecall_sm4_ctr();
    if (ret != 0)
    {
        printf("test evp_sm4_ctr returned error %d\n", ret);
        exit(ret);
    }
    printf("test evp_sm4_ctr completed\n");
}
