/*
 * Copyright (C) 2011-2021 Intel Corporation. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of Intel Corporation nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <sys/time.h>   // for gettimeofday()

#include <unistd.h>
#include <pwd.h>
#include <libgen.h>
#include <stdlib.h>
#include <pthread.h>

#include <openssl/bn.h>
#include <openssl/rand.h>
#include <time.h>

# define MAX_PATH FILENAME_MAX
//# define LENGTH 1048576 // read length in bytes
//# define hashSize 35
//# define rSize 33


#include <sgx_urts.h>

#include "TestApp.h"

#include "TestEnclave_u.h"



/* Global EID shared by multiple threads */
sgx_enclave_id_t global_eid = 0;

typedef struct _sgx_errlist_t {
    sgx_status_t err;
    const char *msg;
    const char *sug; /* Suggestion */
} sgx_errlist_t;

/* Error code returned by sgx_create_enclave */
static sgx_errlist_t sgx_errlist[] = {
    {
        SGX_ERROR_UNEXPECTED,
        "Unexpected error occurred.",
        NULL
    },
    {
        SGX_ERROR_INVALID_PARAMETER,
        "Invalid parameter.",
        NULL
    },
    {
        SGX_ERROR_OUT_OF_MEMORY,
        "Out of memory.",
        NULL
    },
    {
        SGX_ERROR_ENCLAVE_LOST,
        "Power transition occurred.",
        "Please refer to the sample \"PowerTransition\" for details."
    },
    {
        SGX_ERROR_INVALID_ENCLAVE,
        "Invalid enclave image.",
        NULL
    },
    {
        SGX_ERROR_INVALID_ENCLAVE_ID,
        "Invalid enclave identification.",
        NULL
    },
    {
        SGX_ERROR_INVALID_SIGNATURE,
        "Invalid enclave signature.",
        NULL
    },
    {
        SGX_ERROR_OUT_OF_EPC,
        "Out of EPC memory.",
        NULL
    },
    {
        SGX_ERROR_NO_DEVICE,
        "Invalid Intel® Software Guard Extensions device.",
        "Please make sure Intel® Software Guard Extensions module is enabled in the BIOS, and install Intel® Software Guard Extensions driver afterwards."
    },
    {
        SGX_ERROR_MEMORY_MAP_CONFLICT,
        "Memory map conflicted.",
        NULL
    },
    {
        SGX_ERROR_INVALID_METADATA,
        "Invalid enclave metadata.",
        NULL
    },
    {
        SGX_ERROR_DEVICE_BUSY,
        "Intel® Software Guard Extensions device was busy.",
        NULL
    },
    {
        SGX_ERROR_INVALID_VERSION,
        "Enclave version was invalid.",
        NULL
    },
    {
        SGX_ERROR_INVALID_ATTRIBUTE,
        "Enclave was not authorized.",
        NULL
    },
    {
        SGX_ERROR_ENCLAVE_FILE_ACCESS,
        "Can't open enclave file.",
        NULL
    },
};

/* Check error conditions for loading enclave */
void print_error_message(sgx_status_t ret)
{
    size_t idx = 0;
    size_t ttl = sizeof sgx_errlist/sizeof sgx_errlist[0];

    for (idx = 0; idx < ttl; idx++) {
        if(ret == sgx_errlist[idx].err) {
            if(NULL != sgx_errlist[idx].sug)
                printf("Info: %s\n", sgx_errlist[idx].sug);
            printf("Error: %s\n", sgx_errlist[idx].msg);
            break;
        }
    }
    
    if (idx == ttl)
        printf("Error: Unexpected error occurred [0x%x].\n", ret);
}

/* Initialize the enclave:
 *   Step 1: retrive the launch token saved by last transaction
 *   Step 2: call sgx_create_enclave to initialize an enclave instance
 *   Step 3: save the launch token if it is updated
 */
int initialize_enclave(void)
{
    char token_path[MAX_PATH] = {'\0'};
    sgx_launch_token_t token = {0};
    sgx_status_t ret = SGX_ERROR_UNEXPECTED;
    int updated = 0;
    /* Step 1: retrive the launch token saved by last transaction */

    /* try to get the token saved in $HOME */
    const char *home_dir = getpwuid(getuid())->pw_dir;
    if (home_dir != NULL && 
        (strlen(home_dir)+strlen("/")+sizeof(TOKEN_FILENAME)+1) <= MAX_PATH) {
        /* compose the token path */
        strncpy(token_path, home_dir, strlen(home_dir));
        strncat(token_path, "/", strlen("/"));
        strncat(token_path, TOKEN_FILENAME, sizeof(TOKEN_FILENAME)+1);
    } else {
        /* if token path is too long or $HOME is NULL */
        strncpy(token_path, TOKEN_FILENAME, sizeof(TOKEN_FILENAME));
    }

    FILE *fp = fopen(token_path, "rb");
    if (fp == NULL && (fp = fopen(token_path, "wb")) == NULL) {
        printf("Warning: Failed to create/open the launch token file \"%s\".\n", token_path);
    }
    printf("token_path: %s\n", token_path);
    if (fp != NULL) {
        /* read the token from saved file */
        size_t read_num = fread(token, 1, sizeof(sgx_launch_token_t), fp);
        if (read_num != 0 && read_num != sizeof(sgx_launch_token_t)) {
            /* if token is invalid, clear the buffer */
            memset(&token, 0x0, sizeof(sgx_launch_token_t));
            printf("Warning: Invalid launch token read from \"%s\".\n", token_path);
        }
    }

    /* Step 2: call sgx_create_enclave to initialize an enclave instance */
    /* Debug Support: set 2nd parameter to 1 */

    ret = sgx_create_enclave(TESTENCLAVE_FILENAME, SGX_DEBUG_FLAG, &token, &updated, &global_eid, NULL);

    if (ret != SGX_SUCCESS) {
        print_error_message(ret);
        if (fp != NULL) fclose(fp);

        return -1;
    }

    /* Step 3: save the launch token if it is updated */

    if (updated == FALSE || fp == NULL) {
        /* if the token is not updated, or file handler is invalid, do not perform saving */
        if (fp != NULL) fclose(fp);
        return 0;
    }

    /* reopen the file with write capablity */
    fp = freopen(token_path, "wb", fp);
    if (fp == NULL) return 0;
    size_t write_num = fwrite(token, 1, sizeof(sgx_launch_token_t), fp);
    if (write_num != sizeof(sgx_launch_token_t))
        printf("Warning: Failed to save launch token to \"%s\".\n", token_path);
    fclose(fp);

    return 0;
}

/* OCall functions */
void uprint(const char *str)
{
    /* Proxy/Bridge will check the length and null-terminate 
     * the input string to prevent buffer overflow. 
     */
    printf("%s", str);
    fflush(stdout);
}


void usgx_exit(int reason)
{
	printf("usgx_exit: %d\n", reason);
	exit(reason);
}


void* thread_test_func(void* p)
{
	new_thread_func(global_eid);
	return NULL;
}

int ucreate_thread()
{
	pthread_t thread;
	int res = pthread_create(&thread, NULL, thread_test_func, NULL);
	return res;
}

/* string to hex */
void  str2hex(char* text, char* hex_str)
{
    int loop;
    int i; 
    
    i=0;
    loop=0;
    
    while(text[loop] != '\0')
    {
        sprintf((char*)(hex_str+i),"%02X", text[loop]);
        loop+=1;
        i+=2;
    }
    //insert NULL at the end of the output string
    hex_str[i++] = '\0';
}

/* Application entry */
int main(int argc, char *argv[])
{
    (void)(argc);
    (void)(argv);

    /* Changing dir to where the executable is.*/
    char absolutePath[MAX_PATH];
    char *ptr = NULL;
    int LENGTH = 1048576;
    //clock_t start, end; // used to test time
    //double cpu_time_used; // used to test time
    FILE *coverFp;  // cover file
	FILE *randomFp; // files which is used to store hash and r 
	// char coverBuf[LENGTH + 1];
	// char randomBuf[2306870];
	// char coverBufHex [LENGTH * 2 + 1];
	// //char randomBufHex [LENGTH * 2 + 1];
    // //char randomBufHex [65536][34];
	// coverBuf[LENGTH] = '\0';
	// randomBuf[2306869] = '\0';

    char *coverBuf = (char*)malloc(5 * LENGTH + 1);
    char *randomBuf = (char*)malloc(5 * 2306870 + 1);
    char *coverBufHex = (char*)malloc(LENGTH * 10);
	coverBuf[5 * LENGTH] = '\0';
	randomBuf[5 * 2306870] = '\0';
	int result = 5;
    //int secretLen = LENGTH * 2 + 1;
    int secretLen = 15 * LENGTH;
    long secondsSGX;
    long microsSGX;
    long totalSGX = 0;
	

    ptr = realpath(dirname(argv[0]), absolutePath);

    if (ptr == NULL || chdir(absolutePath) != 0)
    	return 1;

    /* Initialize the enclave */
    if (initialize_enclave() < 0)
        return 1; 
 

    /* evaluate execution time */
    //start = clock();
    struct timeval start, end, startSGX, endSGX;
 
    gettimeofday(&start, NULL);

    /**********************open cover file and secret file*******/
	coverFp = fopen("cover.txt", "r");// open cover file
	if(coverFp == NULL) {
		perror("Open cover file error!\n");
		return 0;
	}

	randomFp = fopen("newHash.txt", "r");// open hash file
	if(randomFp == NULL) {
		perror("Open hash file error!\n");
		return 0;
	}

	time_t now1;
 
    // Obtain current time
    // `time()` returns the current time of the system as a `time_t` value
    time(&now1);
 
    // Convert to local time format and print to stdout
    printf("Today is %s", ctime(&now1));


    /* continue to read cover file and hash file */
    //while(!feof(randomFp)) {
    for(int i = 0; i < 3; i++) {
		fread(coverBuf, 5 * LENGTH, 1, coverFp);
		fread(randomBuf, 5 * 2306868, 1, randomFp);
        //fgets(randomBuf, 1024, randomFp);
		str2hex(coverBuf, coverBufHex);
		//printf("%s\n", randomBuf);
		//BN_hex2bn(&m, coverBufHex);
    	//BN_hex2bn(&rCover, randomBuf);
        char *secretMsg = (char*)malloc(secretLen);
        gettimeofday(&startSGX, NULL);
		calSecret(global_eid, coverBufHex, randomBuf, secretMsg, secretLen);
        gettimeofday(&endSGX, NULL);
        secondsSGX = (endSGX.tv_sec - startSGX.tv_sec);
        microsSGX = ((secondsSGX * 1000000) + endSGX.tv_usec) - (startSGX.tv_usec);
        totalSGX += microsSGX;
        printf("%s\n", secretMsg);
		
	}


    /* destroy the enclave */
    sgx_destroy_enclave(global_eid);

    /* evaluate time ends */
    //end = clock();
	//cpu_time_used = ((double) (end - start)) / CLOCKS_PER_SEC;
    fclose(coverFp);
    fclose(randomFp);
    gettimeofday(&end, NULL);
    long seconds = (end.tv_sec - start.tv_sec);
    long micros = ((seconds * 1000000) + end.tv_usec) - (start.tv_usec);
 
    printf("The elapsed time is  %d micros\n", micros);
    printf("time in sgx is %d micros\n", totalSGX);

	//printf("time needs is  %f seconds to execute \n", cpu_time_used);
    time_t now;
    time(&now);
 
    // Convert to local time format and print to stdout
    printf("Today is %s", ctime(&now));
    
    return 0;
}
